#include "SDL.h"
#include "SDL_mixer.h"
#include <cstdlib>
#include <algorithm>

// * (C)2006 Game ported by MARKUS TRENKWALDER */

using namespace std;

extern "C"
{

bool keys[128] = {false};

Uint8 scanmap[SDLK_LAST] = {0};

static SDL_Surface *screen = 0;
static bool fullscreen = false;

struct LinSprite {
	void*  adr;
	Uint32 w, h;
};

bool InitVesa()
{
	int result = SDL_Init(SDL_INIT_VIDEO);
	if ( result == 0 ) atexit(SDL_Quit);
	SDL_WM_SetCaption("MBM2000", NULL);
	
	FILE* f = fopen("data/scanmap", "rb");
	fread(scanmap, sizeof(scanmap), 1, f);
	fclose(f);
	
	return result == 0;
}

bool InitVesamode(Uint32)
{
	Uint32 flags = SDL_HWPALETTE | (fullscreen ? SDL_FULLSCREEN : 0);
	screen = SDL_SetVideoMode(640, 400, 8, flags);
	SDL_ShowCursor(!fullscreen);
	return screen != 0;
}

void filld(void* dest, unsigned count, char value)
{
	unsigned *d = (unsigned*)dest;
	unsigned v  = value;
	
	for ( int i = 0; i < count; ++i )
		*d++ = v;
}

void moved(const void *src, void *dst, unsigned count)
{
	unsigned *d = (unsigned*)dst;
	unsigned *s = (unsigned*)src;
	
	for ( int i = 0; i < count; ++i )
		*d++ = *s++;
}

static Uint32 now = 0;
static const Uint32 STEPSIZE = 14;
static const Uint32 MAX_ALLOWED_FRAMETIME = 10 * STEPSIZE;

Uint32 ProcessEvents()
{
	SDL_Event e;
	while (SDL_PollEvent(&e)) {
		switch ( e.type ) {
		case SDL_QUIT: exit(0);
		case SDL_KEYDOWN:
			// ALT+F4 quits the application (on every platform and even when
			// using the windib driver)
			if (e.key.keysym.sym == SDLK_F4 && 
			    e.key.keysym.mod & KMOD_LALT &&
			    !(e.key.keysym.mod & KMOD_CTRL))
			{
				exit(0);
			}
			
			if ( e.key.keysym.sym == SDLK_RETURN && e.key.keysym.mod & KMOD_ALT ) {
				SDL_Color palette[256];
				memcpy(palette, screen->format->palette->colors,
					sizeof(SDL_Color) * 256);
				fullscreen = !fullscreen;
				Uint32 flags  = SDL_HWPALETTE;
					   flags |= fullscreen ? SDL_FULLSCREEN : 0;
				screen = SDL_SetVideoMode(640, 400, 8, flags);
				SDL_SetPalette(screen, SDL_PHYSPAL | SDL_LOGPAL, palette, 0, 256);
				SDL_ShowCursor(!fullscreen);
				
				// FIX: When the new video mode is set the SDL_KEYUP event for
				// the currently pressed keys seems to always return a scancode
				// of 0 (SDL 1.2.10), which would make the game belive some keys
				// are still pressed. So tell the game that no key is pressed.
				for ( int i = 0; i < 128; ++i ) keys[i] = false;
				
				break;
			}
			/* fall through */
		case SDL_KEYUP: {
			const Uint8 scancode = scanmap[e.key.keysym.sym];
			if ( scancode < 128 ) keys[scancode] = e.type == SDL_KEYDOWN;
			}
		}
	}
	
	Uint32 count;

	do
	{
		Uint32 cur = SDL_GetTicks();
		count = 0;
			
		if( cur-now > MAX_ALLOWED_FRAMETIME ) now = cur-STEPSIZE;

		count = (cur - now) / STEPSIZE;
		now += count * STEPSIZE;

		// This caps the framerate to 70 fps and thus makes
		// the game less cpu intensive.
		if ( count == 0 ) SDL_Delay(STEPSIZE - (cur - now));
	}
	while ( count == 0 );
	
	return count;
}


void setpalette(const Uint8 *pal)
{
	SDL_Color colors[256];
	for ( int i = 0; i < 256; ++i ) {
		colors[i].r = *pal++;
		colors[i].g = *pal++;
		colors[i].b = *pal++;
	}
	int rs = SDL_SetPalette(screen, SDL_PHYSPAL | SDL_LOGPAL, colors, 0, 256);
}

void decpalette(Uint8 num, Uint8 *pal)
{
	for ( int i = 0; i < 3 * 256; ++i )		
		pal[i] = max(pal[i] - num, 0);
	setpalette(pal);
}

void waitretrace()
{
}

void CopyScreen100(const Uint32* data)
{
	if ( SDL_MUSTLOCK(screen) )
		while ( SDL_LockSurface(screen) != 0 )
			SDL_Delay(10);
			
	moved(data, (Uint32*)screen->pixels, 64000);
	
	if ( SDL_MUSTLOCK(screen) )
		SDL_UnlockSurface(screen);
		
	SDL_Flip(screen);
}

void Putsolidsprt100(void* dest, Sint32 x, Sint32 y, const LinSprite* s)
{
	x -= s->w >> 1;
	y -= s->h >> 1;	
	Sint32 r  = x + s->w;
	Sint32 b  = y + s->h;
	
	Uint8 *dst = (Uint8*)dest;
	Uint8 *src = (Uint8*)s->adr;
	Uint32 dst_skip;
	Uint32 src_skip = 0;
	
	if ( x < 0 )   { src_skip += -x; src += -x; x = 0; }
	if ( r > 640 ) { src_skip += r - 640; r = 640;}
	if ( y < 0 )   { src += -y * s->w; y = 0; }
	if ( b > 400 ) { b = 400; }
	
	Sint32 dtx = r - x;
	Sint32 dty = b - y;
	
	dst_skip = 640 - dtx;
	dst += x + y * 640;
	
	for ( int i = 0; i < dty; ++i ) {
		for ( int j = 0; j < dtx; ++j )
			*dst++ = *src++;
		dst += dst_skip;
		src += src_skip;		
	}
}

void Putsprt100(void* dest, Sint32 x, Sint32 y, const LinSprite* s)
{
	x -= s->w >> 1;
	y -= s->h >> 1;	
	Sint32 r  = x + s->w;
	Sint32 b  = y + s->h;
	
	Uint8 *dst = (Uint8*)dest;
	Uint8 *src = (Uint8*)s->adr;
	Uint32 dst_skip;
	Uint32 src_skip = 0;
	
	if ( x < 0 )   { src_skip += -x; src += -x; x = 0; }
	if ( r > 640 ) { src_skip += r - 640; r = 640;}
	if ( y < 0 )   { src += -y * s->w; y = 0; }
	if ( b > 400 ) { b = 400; }
	
	Sint32 dtx = r - x;
	Sint32 dty = b - y;
	
	dst_skip = 640 - dtx;
	dst += x + y * 640;
	
	for ( int i = 0; i < dty; ++i ) {
		for ( int j = 0; j < dtx; ++j )
			if ( *src ) *dst++ = *src++;
			else dst++, src++;
		dst += dst_skip;
		src += src_skip;		
	}
}

void Putscalsprt100(void* dest, Sint32 x, Sint32 y, Uint32 sx, Uint32 sy, const LinSprite* s)
{
	const Uint32 w = (s->w * sx) / 100;
	const Uint32 h = (s->h * sy) / 100;
	x -= w >> 1;
	y -= h >> 1;	
	Sint32 r  = min(x + int(w), 640);
	Sint32 b  = min(y + int(h), 400);	
	
	Uint8 *dst = (Uint8*)dest;
	Uint8 *src = (Uint8*)s->adr;
	Uint32 dst_skip;
	
	Uint32 start_sdx = 0;
	Uint32 start_skip = 0;
	Uint32 sdy = 0;
	
	if ( x < 0 ) {
		div_t r = div(-x * 100, sx);
		start_skip = r.quot;
		start_sdx  = r.rem;
		x = 0;
	}
	if ( y < 0 ) {
		div_t r = div(-y * 100, sy);
		sdy = r.rem;
		src += r.quot * s->w;
		y = 0;
	}
	
	Sint32 dtx = r - x;
	Sint32 dty = b - y;
	
	dst_skip = 640 - dtx;
	dst += x + y * 640;
	
	for ( int i = 0; i < dty; ++i ) {
		Uint8 *srcc = src + start_skip;
		Uint8 color = *srcc++;
		Uint32 sdx = start_sdx;
		for ( int j = 0; j < dtx; ++j ) {
			if ( color ) *dst++ = color;
			else dst++;
			
			sdx += 100;
			while ( sdx >= sx ) {
				color = *srcc++;
				sdx -= sx;
			}
		}
		dst += dst_skip;
		sdy += 100;
		while ( sdy >= sy ) {
			src += s->w;
			sdy -= sy;
		}
	}
}

void Putdotscalsprt100(void *dest, Sint32 x, Sint32 y, Uint32 sx, Uint32 sy, const LinSprite* s)
{
	const Uint32 w = (s->w * sx) / 100;
	const Uint32 h = (s->h * sy) / 100;
	x -= w >> 1;
	y -= h >> 1;	
	Sint32 r  = min(x + int(w), 640);
	Sint32 b  = min(y + int(h), 400);	
	
	Uint8 *dst = (Uint8*)dest;
	Uint8 *src = (Uint8*)s->adr;
	Uint32 dst_skip;
	
	Uint32 start_sdx = 0;
	Uint32 start_skip = 0;
	Uint32 sdy = 0;
	
	if ( x < 0 ) {
		div_t r = div(-x * 100, sx);
		start_skip = r.quot;
		start_sdx  = r.rem;
		x = 0;
	}
	if ( y < 0 ) {
		div_t r = div(-y * 100, sy);
		sdy = r.rem;
		src += r.quot * s->w;
		y = 0;
	}
	
	Sint32 dtx = r - x;
	Sint32 dty = b - y;
	
	dst_skip = 640 - dtx;
	dst += x + y * 640;
	
	for ( int i = 0; i < dty; ++i ) {
		Uint8 *srcc = src + start_skip;
		Uint8 color = *srcc++;
		Uint32 sdx = start_sdx;
		for ( int j = 0; j < dtx; ++j ) {
			if ( color ) *dst++ = color;
			else dst++;
			color = 0;
		
			sdx += 100;
			while ( sdx >= sx ) {
				color = *srcc++;
				sdx -= sx;
			}
		}
		dst += dst_skip;
		while ( (sdy += 100) < sy ) { dst += 640; --dty; }
		while (sdy >= sy) {	src += s->w; sdy -= sy;	}
	}
}

void Line100(Uint8* dst, int x0, int y0, int x1, int y1, Uint8 col)
{
     bool steep = abs(y1 - y0) > abs(x1 - x0);
     if ( steep ) {
         swap(x0, y0);
         swap(x1, y1);
	 }
     if ( x0 > x1 ) {
         swap(x0, x1);
         swap(y0, y1);
	 }
     int deltax = x1 - x0;
     int deltay = abs(y1 - y0);
     int error = 0;
     int y = y0;
	 int ystep = y0 < y1 ? 1 : -1;
	 for ( int x = x0; x <= x1; ++x ) {
         if (steep) dst[y + x * 640] = col;
		 else dst[x + y * 640] * col;
         error = error + deltay;
         if ( 2 * error >= deltax ) {
             y += ystep;
             error -= deltax;
		 }
	 }
}

/*--------------------- SDL_mixer bindings -----------------------------------*/

bool Reset_SB()
{
	bool success;
	success = (SDL_InitSubSystem(SDL_INIT_AUDIO) == 0);
    success = success && 
		(Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 1, 2048) == 0);
	return success;
}

bool InitSB(Uint8 no)
{
	return Mix_AllocateChannels(no) == no;
}

void DoneSB()
{
	if ( SDL_WasInit(SDL_INIT_AUDIO) != 0 )  {
		Mix_CloseAudio();
		SDL_QuitSubSystem(SDL_INIT_AUDIO);
	}
}

void ReadWave(const char* filename, Mix_Chunk** wave)
{
	*wave = Mix_LoadWAV(filename);
}

void PutChannel(Mix_Chunk* wav, Uint8 vol, Uint8 ch_no)
{
	Mix_Volume(ch_no - 1, vol * 2);
	Mix_PlayChannel(ch_no - 1, wav, 0);	
}

} // extern "C"
